#pragma once

#include <cstdint>

namespace euler::syscall {

  typedef uint32_t encoded_color;
  typedef int32_t exit_code;
  typedef uint64_t file_handle;
  typedef uint32_t key_packet;

  enum [[nodiscard]] file_result : uint64_t {
    success,
    bad_file_handle,
    device_error,
    file_system_corrupt,
    out_of_bounds,
    does_not_exist,
    directory
  };

}

extern "C" {

  euler::syscall::encoded_color _syscall_encode_color(
    uint8_t red, uint8_t green, uint8_t blue);

  void _syscall_get_framebuffer(
    euler::syscall::encoded_color *&ptr_out, uint32_t &width_out,
    uint32_t &height_out, uint32_t &pitch_out);

  euler::syscall::file_result _syscall_open_file(
    const char *path, uint64_t path_length, euler::syscall::file_handle &out);

  euler::syscall::file_result _syscall_get_file_length(
    euler::syscall::file_handle file, uint64_t &out);

  euler::syscall::file_result _syscall_read_from_file(
    euler::syscall::file_handle file,
    uint64_t start_offset, uint64_t length, void *into);

  [[noreturn]] void _syscall_end_this_process(euler::syscall::exit_code code);

  [[nodiscard]] void *_syscall_get_new_pages(uint64_t count);

  void _syscall_close_file(euler::syscall::file_handle file);

  euler::syscall::key_packet _syscall_read_key_packet();

}
